;; x-message.jl -- Slightly fancier message display

;; version 0.2

;; Copyright (C) 2000 merlin <merlin@merlin.org>

;; this is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; this is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with sawmill; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

(provide 'x-message)

(require 'x)

(defvar message-window nil)
(defvar message-gc nil)
(defvar message-msg nil)
(defvar message-attrs nil)
(defvar message-pos (cons 0 0))
(defvar message-dims (cons 0 0))

(defvar default-message-padding (cons 4 4))
(defvar default-message-foreground "black")
(defvar default-message-background "white")
(defvar default-message-border-color "black")
(defvar default-message-border-width 1)
(defvar default-message-spacing 1)
(defvar default-message-position (cons (quotient (screen-width) 2)  (quotient (screen-height) 2)))

(defun assqd (key alist default)
  (if (assq key alist)
    (assq key alist)
    (cons key default)))

(defun fontify (font)
  (if (stringp font) (get-font font) font))

(defun colorify (color)
  (if (stringp color) (get-color color) color))

(defun x-message-repaint-window (id)
  (when (eq id message-window)
    (let
        ((pad (cdr (assqd 'padding message-attrs default-message-padding)))
	 (fg (colorify (cdr (assqd 'foreground message-attrs default-message-foreground))))
	 (font (fontify (cdr (assq 'font message-attrs))))
	 (justify (cdr (assqd 'x-justify message-attrs 'left)))
	 (spacing (cdr (assqd 'spacing message-attrs default-message-spacing)))
         (w (car message-dims)) x y)
     (setq y (cdr pad))
     (x-clear-window message-window)
     (x-change-gc message-gc `((foreground . ,fg)))
     (mapcar (lambda (msg)
         (when (stringp msg)
           (cond ((eq 'left justify)
		  (setq x (car pad)))
		 ((eq 'center justify)
		  (setq x (quotient (- w (text-width msg font)) 2)))
		 (t ;; (eq 'right justify)
		  (setq x (- w (text-width msg font) (car pad)))))
           (setq y (+ y (font-ascent font) spacing)) ;; spacing not on first line!
           (x-draw-string message-window message-gc (cons x y) msg font)
           (setq y (+ y (font-descent font))))
         (when (consp msg)
           (cond ((eq 'font (car msg))
		  (setq font (fontify (cdr msg))))
		 ((eq 'foreground (car msg))
		  (x-change-gc message-gc `((foreground . ,(colorify (cdr msg))))))
		 ((eq 'x-justify (car msg))
		  (setq justify (cdr msg)))
		 ((eq 'spacing (car msg))
		  (setq spacing (cdr msg))))))
          message-msg))))

(defun x-message-calculate-dimensions ()
  (let
      ((pad (cdr (assqd 'padding message-attrs default-message-padding)))
       (font (fontify (cdr (assq 'font message-attrs))))
       (spacing (cdr (assqd 'spacing message-attrs default-message-spacing))))
     (setq message-dims (cons (* 2 (car pad)) (* 2 (cdr pad))))
     (mapcar (lambda (msg)
         (when (stringp msg)
           (rplaca message-dims
                (max (car message-dims) (+ (* 2 (car pad)) (text-width msg font))))
           (rplacd message-dims
                (+ (cdr message-dims) spacing (font-height font)))) ;; spacing not on first line!
         (when (consp msg)
           (cond ((eq 'font (car msg))
		  (setq font (fontify (cdr msg))))
		 ((eq 'spacing (car msg))
		  (setq spacing (cdr msg))))))
          message-msg)))

(defun x-message-calculate-position ()
  (let
      ((pos (cdr (assqd 'position message-attrs default-message-position)))
       (bw (cdr (assqd 'border-width message-attrs default-message-border-width)))
       (gravity (cdr (assqd 'gravity message-attrs 'center)))
       (w (car message-dims))
       (h (cdr message-dims)) x y)
    (cond ((memq gravity '(north center south))
	   (setq x (- (car pos) (quotient w 2) bw)))
	  ((memq gravity '(north-west west south-west))
	   (setq x (- (car pos) w (* bw 2))))
	  (t (setq x (car pos))))
    (cond ((memq gravity '(west center east))
	   (setq y (- (cdr pos) (quotient h 2) bw)))
	  ((memq gravity '(north-west north north-east))
	   (setq y (- (cdr pos) h (* bw 2))))
	  (t (setq y (cdr pos))))
    (setq message-pos
      (cons (max (min x (- (screen-width) w (* bw 2))) 0)
            (max (min y (- (screen-height) h (* bw 2))) 0)))))

(defun x-message-handler (type &optional args)
  (cond ((eq type 'expose) (x-message-repaint-window message-window))))

(defun x-message-create-window ()
  (let*
      ((bw (cdr (assqd 'border-width message-attrs default-message-border-width)))
       (bg (colorify (cdr (assqd 'background message-attrs default-message-background))))
       (bd (colorify (cdr (assqd 'border-color message-attrs default-message-border-color))))
       (window-attrs `((background . ,bg)
		       (border-color . ,bd)
		       (override-redirect . ,t)
		       (save-under . ,nil)
		       (event-mask . ,'(expose))))
       (gc-attrs `((background . ,bg))))

    (setq message-window (x-create-window message-pos message-dims bw window-attrs x-message-handler))
    (setq message-gc (x-create-gc message-window gc-attrs))
    (x-map-window message-window t)))

(defun x-message-update-window ()
  (let*
      ((x (car message-pos))
       (y (cdr message-pos))
       (w (car message-dims))
       (h (cdr message-dims))
       (bw (cdr (assqd 'border-width message-attrs default-message-border-width)))
       (bg (colorify (cdr (assqd 'background message-attrs default-message-background))))
       (bd (colorify (cdr (assqd 'border-color message-attrs default-message-border-color))))
       (window-config `((x . ,x) (y  . ,y)
			(width . ,w) (height . ,h)
			(border-width . ,bw)
			(stack-mode . top-if)))
       (window-attrs `((background . ,bg)6
		       (border-color . ,bd)))
       (gc-attrs `((background . ,bg))))

    (x-configure-window message-window window-config)
    (x-change-window-attributes message-window window-attrs)
    (x-change-gc message-gc gc-attrs)))

(defun x-message-hide ()
  (when message-window
    (x-destroy-window message-window)
    (setq message-window nil))
  (when message-gc
    (x-destroy-gc message-gc)
    (setq message-gc nil)))

;; supported global attributes:
;;
;; 'position - (x . y) position
;; 'gravity - how the window is positioned relative to position
;; 'font - default font
;; 'foreground - default foreground
;; 'background - default background
;; 'border-color - border color
;; 'font - default font
;; 'x-justify - default justification
;; 'spacing - interline spacing
;; 'padding - (x . y) outer padding
;; 'border-width - border width

;; supported inline attributes:
;;
;; 'font - font
;; 'foreground - foreground
;; 'x-justify - justification
;; 'spacing - interline spacing

(defun x-message-display (message attrs)
  (setq message-msg message)
  (setq message-attrs attrs)
  (x-message-calculate-dimensions)
  (x-message-calculate-position)
  (if message-window
    (x-message-update-window)
    (x-message-create-window))
  (x-message-repaint-window message-window))

;; just a hack to see about overriding the default display-message

;; 'font 'background 'foreground 'x-justify 'spacing 'position

;(defun split-nl (msg)
;  (let
;    ((index 0)
;     (split nil))
;    (while
;      (string-looking-at "([ -z]+)[^ -z]?" msg index)
;      (setq split (nconc split (list (substring msg index (match-end 1)))))
;      (setq index (match-end)))
;    split))

;(defun zoo-display-message (msg &optional attrs)
;  (if msg
;    (let ((pos (cdr (assq 'position attrs))) gravity)
;      (when pos
;        (cond
;          ((and (>= (car pos) 0) (>= (cdr pos) 0))
;           (setq gravity 'south-east))
;          ((and (< (car pos) 0) (>= (cdr pos) 0))
;           (setq gravity 'south-west)
;           (rplaca pos (+ (screen-width) (car pos))))
;          ((and (< (car pos) 0) (< (cdr pos) 0))
;           (setq gravity 'north-west)
;           (rplaca pos (+ (screen-width) (car pos)))
;           (rplacd pos (+ (screen-height) (cdr pos))))
;          ((and (>= (car pos) 0) (< (cdr pos) 0))
;           (setq gravity 'north-east)
;           (rplacd pos (+ (screen-height) (cdr pos)))))
;        (setq attrs (nconc attrs `((gravity . ,gravity)))))
;      (x-message-display (split-nl msg) attrs))
;    (x-message-hide)))

;; bugs: spacing is added before the first line
